local json = require("json")
local m = {}

function GetVersion()
  return 2
end

local dpUrnTable = 
{
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-1:Value"},   -- DLM: application parameter  D1: activated state	NP: I1 B1 Bit2	
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-15:Value"},  -- DLM: application parameter  D7:	overcurrent protection 16,20,25,32,35,40,50,63,80,100,125,160,200,224,250 in A	NP: I15
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-16:Value"},  -- DLM: application parameter  D6:	load reserve 0…5….100 in %  NP: I16
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-17:Value"},  -- DLM: application parameter  D10: prioritised charge point 0…5	0 => unused.	NP: I17
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-19:Value"},  -- grid serving interruption   A2: setpoint value as ampere 0 or 6 NP: I19
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-2:Value"},   -- charge points data:         A1: charge point count 1..5 NP: I2
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-1:Value"},     -- charge points data: charge point 1 manufacturer TP: I1
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-2:Value"},     -- charge points data: charge point 1 model        TP: I2
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-3:Value"},     -- charge points data: charge point 2 manufacturer TP: I3
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-4:Value"},     -- charge points data: charge point 2 model        TP: I4
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-5:Value"},     -- charge points data: charge point 3 manufacturer TP: I5
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-6:Value"},     -- charge points data: charge point 3 model        TP: I6
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-7:Value"},     -- charge points data: charge point 4 manufacturer TP: I7
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-8:Value"},     -- charge points data: charge point 4 model        TP: I8
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-9:Value"},     -- charge points data: charge point 5 manufacturer TP: I9
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-10:Value"},    -- charge points data: charge point 5 model        TP: I10
  {urn = "urn:gds:dp:ECHIKM01:Co-81:Value"},    -- charge points data: charge point 1 Firmwareversion   CO: 81
  {urn = "urn:gds:dp:ECHIKM01:Co-131:Value"},   -- charge points data: charge point 2 Firmwareversion   CO: 131
  {urn = "urn:gds:dp:ECHIKM01:Co-181:Value"},   -- charge points data: charge point 3 Firmwareversion   CO: 181
  {urn = "urn:gds:dp:ECHIKM01:Co-231:Value"},   -- charge points data: charge point 4 Firmwareversion   CO: 231
  {urn = "urn:gds:dp:ECHIKM01:Co-281:Value"},   -- charge points data: charge point 4 Firmwareversion   CO: 281
}

local dpUrnSmartMeterTable = 
{
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-26:Value"},  -- charge points data: charge point 1 L13: activated external SmartMeter NP:26 B2 Bit1
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-56:Value"},  -- charge points data: charge point 2 L13: activated external SmartMeter NP:56 B2 Bit1
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-86:Value"},  -- charge points data: charge point 3 L13: activated external SmartMeter NP:86 B2 Bit1
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-116:Value"},  -- charge points data: charge point 4 L13: activated external SmartMeter NP:116 B2 Bit1
  {urn = "urn:gds:dp:ECHIKM01:Number-Parameter-Object-146:Value"},  -- charge points data: charge point 5 L13: activated external SmartMeter NP:146 B2 Bit1
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-21:Value"},     -- charge points data: charge point 1 SmartMeter name and type TP: I21
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-22:Value"},     -- charge points data: charge point 2 SmartMeter name and type TP: I22
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-23:Value"},     -- charge points data: charge point 3 SmartMeter name and type TP: I23
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-24:Value"},     -- charge points data: charge point 4 SmartMeter name and type TP: I24
  {urn = "urn:gds:dp:ECHIKM01:Text-Parameter-Object-25:Value"},     -- charge points data: charge point 5 SmartMeter name and type TP: I25
}

function GetCommands()
  return json.encode(
    {
      {
        type = "function",
        func = "GetUptime"
      },
      {
        type = "api",
        -- devicestack state: send delay if too long start-up time is detected
        data = json.encode({command = "getdevicestate"}),
        func = "HandleGetDeviceState"
      },
      {
        type = "api",
        -- default device configuration data
        data = json.encode({command = "getdeviceconfig", ipc = true}),
        func = "HandleGetDeviceConfig"
      },
      {
        type = "api",
        -- KNX Application program loaded state
        data = json.encode({command = "getconfiguration", object = {urn = "urn:gds:chn:ECHIKM01:EChargeIIApp"}}),
        func = "HandleGetApplicationLoadStateConfiguration"
      },
      {
        type = "api",
        -- grid serving interruption: linked status from communication data point
        data = json.encode({command = "getconfiguration", object = {urn = "urn:gds:dp:ECHIKM01:Co-1:Value"}}),
        func = "HandleGetCo1Configuration"
      },
      {
        type = "api",
        -- all values in dpUrnTable table
        data = json.encode({command = "getvalues", values = dpUrnTable, internal = true}),
        func = "HandleGetValues"
      },
      {
        type = "api",
        -- all values in dpUrnSmartMeterTable table
        data = json.encode({command = "getvalues", values = dpUrnSmartMeterTable, internal = true}),
        func = "HandleGetValues"
      },
      {
        type = "api",
        -- DLM data: peak consumers used CO13: urn = urn:gds:dp:ECHIKM01:Co-13:Value
        data = json.encode({command = "getconfiguration", object = {urn = "urn:gds:dp:ECHIKM01:Co-13:Value"}}),
        func = "HandleGetCo13Configuration"
      },
      {
        -- eCharge: collect e-charge specific data
        type = "function",
        func = "CreateEChargeData"
      }
    }
  )
end

------ local definitions ------

-- When next is local, the code finds primitive function next by a constant-time indexing operation into an array of "upvalues."
local next = next

local eChargeTable = {}

local applicationLoadState = true
m.APPLICATION_LOAD_STATE_KEY_NAME = "appLoaded"
-- Table for the grid serving interruption data
local gridServingTable = {}
m.GRID_SERVING_TABLE_NAME = "gridServingInterruption"
m.GRID_SERVING_TABLE_KEY_USED = "used"
m.GRID_SERVING_TABLE_KEY_SETPOINT = "setpoint"

-- Load management status NP: I1 B1 Bit2	- 0 => deactivated, 1 => activated
local lmActivated = true
-- Load management type NP:I1 B3 Bit 4 - 0 => DLM, 1 => SLM
m.LMTYPE_DLM = 0
m.LMTYPE_SLM = 1
m.LMTYPE = m.LMTYPE_DLM

m.LM_TABLE_ACTIVATED_INDEX = 0
m.LM_TABLE_OVERCURRENT_INDEX = 1
m.LM_TABLE_LOAD_RESERVE_INDEX = 2
m.LM_TABLE_PRIORITISED_CHARGINGPOINT_INDEX = 3
-- CO 13
m.LM_TABLE_PEAK_CONSUMERS_INDEX = 4

local lmTableKeyName = {}
lmTableKeyName[m.LMTYPE_DLM] = {}
lmTableKeyName[m.LMTYPE_SLM] = {}
-- DLM
lmTableKeyName[m.LMTYPE_DLM][m.LM_TABLE_ACTIVATED_INDEX] = "activated"
lmTableKeyName[m.LMTYPE_DLM][m.LM_TABLE_OVERCURRENT_INDEX] = "overcurrentProtection"
lmTableKeyName[m.LMTYPE_DLM][m.LM_TABLE_LOAD_RESERVE_INDEX] = "loadReserve"
lmTableKeyName[m.LMTYPE_DLM][m.LM_TABLE_PRIORITISED_CHARGINGPOINT_INDEX] = "prioritisedChargePointUsed"
-- CO 13
lmTableKeyName[m.LMTYPE_DLM][m.LM_TABLE_PEAK_CONSUMERS_INDEX] = "peakConsumersUsed"
-- SLM
lmTableKeyName[m.LMTYPE_SLM][m.LM_TABLE_ACTIVATED_INDEX] = "activated"
lmTableKeyName[m.LMTYPE_SLM][m.LM_TABLE_OVERCURRENT_INDEX] = "maxChargingCurrent"
lmTableKeyName[m.LMTYPE_SLM][m.LM_TABLE_LOAD_RESERVE_INDEX] = "loadReserve"
lmTableKeyName[m.LMTYPE_SLM][m.LM_TABLE_PRIORITISED_CHARGINGPOINT_INDEX] = "prioritisedChargePointUsed"

-- Table for the DLM data
m.DLM_TABLE_NAME = "dlm"
local dlmTable = {}
local dlmActivated = true

-- Table name for the SLM data
m.SLM_TABLE_NAME = "slm"
local slmTable = {}
local slmActivated = true

-- Table for the charge points
m.CPS_TABLE_NAME = "chargePoints"
local chargePointsTable = {}
m.ECHARGE_TABLE_KEY_COUNT = "chargePointCount"

-- Table for the charge point data
m.CP_DATA_KEY_MANUFACTURER = "manufacturer"
m.CP_DATA_KEY_MODEL = "model"
m.CP_DATA_KEY_FIRMWARE = "firmwareVersion"

-- Table for the charge point smart meter data
m.CP_SMARTMETER_TABLE_NAME = "electricityMeter"
m.CP_SMARTMETER_TABLE_KEY_ACTIVATED = "activated"
m.CP_SMARTMETER_TABLE_KEY_METER = "meter"

--[[
Function to get the number from string. If string is empty then returns nil.
]]
local function GetNumberFromString(x)
  local result = nil
  if x then
    if type(x) == 'number' then return x end
    if type(x) == 'string' and string.len(x) > 0 then 
        result = tonumber(x) 
    end
  end
  return result
end

--[[
Function to set the charge point manufacturer in table chargePointsTable at given index
]]
local function SetChargePointManufacturer(index, manufacturer) 
  if manufacturer then
    if type(manufacturer) ~= 'string' then manufacturer = tostring(manufacturer) end
    if string.len(manufacturer) == 0 then
      return
    end
    
    local key = m.CP_DATA_KEY_MANUFACTURER
    if chargePointsTable[index] == nil then
      chargePointsTable[index] = {}
    end
    chargePointsTable[index][key] = manufacturer   -- charge point manufacturer TP: I1
  end
end

--[[
Function to set the charge point model in table chargePointsTable at given index
]]
local function SetChargePointModel(index, model) 
  if model then
    if type(model) ~= 'string' then model = tostring(model) end
    if string.len(model) == 0 then
      return
    end
    local key = m.CP_DATA_KEY_MODEL
    if chargePointsTable[index] == nil then
      chargePointsTable[index] = {}
    end
    chargePointsTable[index][key] = model   -- charge point model        TP: I2
  end
end

--[[
Function to set the charge point firmware in table chargePointsTable at given index
]]
local function SetChargePointFirmware(index, firmware) 
  if firmware then
    if type(firmware) ~= 'string' then firmware = tostring(firmware) end
    if string.len(firmware) == 0 then
      firmware = "n/a" -- => not available
    end
    local key = m.CP_DATA_KEY_FIRMWARE
    if chargePointsTable[index] == nil then
      -- ignore empty version
    else
      chargePointsTable[index][key] = firmware -- charge point firmware version
    end
  end
end

--[[
Function to set the charge point smart meter state in table chargePointsTable at given index
]]
local function SetSmartMeterActiveState(index, activeState) 
  if activeState ~= nil then
    if chargePointsTable[index] == nil then
      -- ignore empty table
    else
      local keyCp = m.CP_SMARTMETER_TABLE_NAME
      if chargePointsTable[index][keyCp] == nil then
        chargePointsTable[index][keyCp] = {}
      end
      local key = m.CP_SMARTMETER_TABLE_KEY_ACTIVATED
      chargePointsTable[index][keyCp][key] = activeState -- charge point smart meter active state
    end
  end
end

--[[
Function to set the charge point smart meter manufacturer and model in table chargePointsTable at given index
]]
local function SetSmartMeterManufacturer(index, meter) 
  if meter ~= nil then
    if type(meter) ~= 'string' then meter = tostring(meter) end
    if string.len(meter) == 0 then
      return
    end
    if chargePointsTable[index] == nil then
      -- ignore empty table
    else
      local keyCp = m.CP_SMARTMETER_TABLE_NAME
      if chargePointsTable[index][keyCp] == nil then
        chargePointsTable[index][keyCp] = {}
      end
      local key = m.CP_SMARTMETER_TABLE_KEY_METER
      chargePointsTable[index][keyCp][key] = meter -- charge point smart meter manufacturer 
    end
  end
end

-- Function to get the load management type -- D11: NP:I1 B3 Bit 4  (0x100 => 256): 0 => DLM, 1 => SLM
local function GetLoadManagementType(x)
  if (x & 256) > 0 then 
    return m.LMTYPE_SLM 
  else
    return m.LMTYPE_DLM 
  end
end

-- Function to set load management table raw value
local function SetLoadManagementTableRawValue(table, tableLmType, tableKeyIndex, value)
  local keyName = lmTableKeyName[tableLmType][tableKeyIndex]
  if keyName then
    table[keyName] = value
  end
end

-- Function to set load management table data
local function SetLoadManagementTableData(tableKeyIndex, value, checkActivation)
  local activeTable = dlmTable
  local isActive = true
  if m.LMTYPE == m.LMTYPE_SLM then
    activeTable = slmTable
    if checkActivation then
      isActive = slmActivated
    end
  else
    if checkActivation then
      isActive = dlmActivated
    end
  end
  
  if isActive then
    SetLoadManagementTableRawValue(activeTable, m.LMTYPE, tableKeyIndex, value)
  end
end

-- used like a switch case for configuration from getValues response
local dpUrnToValueAction = 
{
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-1:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        lmActivated = (x & 536870912) > 0  -- D1: NP: I1 B1 Bit2	(0x20000000 => 536870912)
        if not lmActivated then
          dlmActivated = false
          slmActivated = false
        else
          -- D11: NP:I1 B3 Bit 4  (0x100 => 256): 0 => DLM, 1 => SLM
          m.LMTYPE = GetLoadManagementType(x)
          if m.LMTYPE == m.LMTYPE_SLM then
            slmActivated = true
            dlmActivated = false
          else
            slmActivated = false
            dlmActivated = true
          end
        end
        SetLoadManagementTableRawValue(dlmTable, m.LMTYPE_DLM, m.LM_TABLE_ACTIVATED_INDEX, dlmActivated)
        SetLoadManagementTableRawValue(slmTable, m.LMTYPE_SLM, m.LM_TABLE_ACTIVATED_INDEX, slmActivated)
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-15:Value"] = 
    function(x) 
      x = GetNumberFromString(x)  -- D7:	16,20,25,32,35,40,50,63,80,100,125,160,200,224,250 A	NP: I15
      if x then
        SetLoadManagementTableData(m.LM_TABLE_OVERCURRENT_INDEX, x, true)
      end
     
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-16:Value"] = 
    function(x) 
      x = GetNumberFromString(x)  -- D6: 0…5….100 %	NP: I16
      if x then
        SetLoadManagementTableData(m.LM_TABLE_LOAD_RESERVE_INDEX, x, true)
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-17:Value"] = 
    function(x) 
      x = GetNumberFromString(x)  -- D10:	0…5	NP: I17
      if x then
        SetLoadManagementTableData(m.LM_TABLE_PRIORITISED_CHARGINGPOINT_INDEX, x > 0, true)
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-19:Value"] = 
    function(x) 
      x = GetNumberFromString(x)
      if x then
        gridServingTable[m.GRID_SERVING_TABLE_KEY_SETPOINT] = x   -- A2:	0 or 6	NP: I19
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-2:Value"] = 
    function(x) 
      x = GetNumberFromString(x)
      if x then
        eChargeTable[m.ECHARGE_TABLE_KEY_COUNT] = x   -- A1:	1..5  NP: I2
      end
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-1:Value"] = 
    function(x) 
      SetChargePointManufacturer(1, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-2:Value"] = 
    function(x) 
      SetChargePointModel(1, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-3:Value"] = 
    function(x) 
      SetChargePointManufacturer(2, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-4:Value"] = 
    function(x) 
      SetChargePointModel(2, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-5:Value"] = 
    function(x) 
      SetChargePointManufacturer(3, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-6:Value"] = 
    function(x) 
      SetChargePointModel(3, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-7:Value"] = 
    function(x) 
      SetChargePointManufacturer(4, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-8:Value"] = 
    function(x) 
      SetChargePointModel(4, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-9:Value"] = 
    function(x) 
      SetChargePointManufacturer(5, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-10:Value"] = 
    function(x) 
      SetChargePointModel(5, x)
    end,
  ["urn:gds:dp:ECHIKM01:Co-81:Value"] = 
    function(x) 
      SetChargePointFirmware(1, x)
    end,
  ["urn:gds:dp:ECHIKM01:Co-131:Value"] = 
    function(x) 
      SetChargePointFirmware(2, x)
    end,
  ["urn:gds:dp:ECHIKM01:Co-181:Value"] = 
    function(x) 
      SetChargePointFirmware(3, x)
    end,
  ["urn:gds:dp:ECHIKM01:Co-231:Value"] = 
    function(x) 
      SetChargePointFirmware(4, x)
    end,
  ["urn:gds:dp:ECHIKM01:Co-281:Value"] = 
    function(x) 
      SetChargePointFirmware(5, x)
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-26:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        local activeState = (x & 4194304) > 0  -- L13: activated external SmartMeter NP:26 B2 Bit1 (0x00400000 => 4194304)
        SetSmartMeterActiveState(1, activeState) 
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-56:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        local activeState = (x & 4194304) > 0  -- L13: activated external SmartMeter NP:56 B2 Bit1 (0x00400000 => 4194304)
        SetSmartMeterActiveState(2, activeState) 
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-86:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        local activeState = (x & 4194304) > 0  -- L13: activated external SmartMeter NP:86 B2 Bit1 (0x00400000 => 4194304)
        SetSmartMeterActiveState(3, activeState) 
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-116:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        local activeState = (x & 4194304) > 0  -- L13: activated external SmartMeter NP:116 B2 Bit1 (0x00400000 => 4194304)
        SetSmartMeterActiveState(4, activeState) 
      end
    end,
  ["urn:gds:dp:ECHIKM01:Number-Parameter-Object-146:Value"] = 
    function(x)
      x = GetNumberFromString(x)
      if x then
        local activeState = (x & 4194304) > 0  -- L13: activated external SmartMeter NP:146 B2 Bit1 (0x00400000 => 4194304)
        SetSmartMeterActiveState(5, activeState) 
      end
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-21:Value"] =
    function(x) 
      SetSmartMeterManufacturer(1, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-22:Value"] =
    function(x) 
      SetSmartMeterManufacturer(2, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-23:Value"] =
    function(x) 
      SetSmartMeterManufacturer(3, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-24:Value"] =
    function(x) 
      SetSmartMeterManufacturer(4, x)
    end,
  ["urn:gds:dp:ECHIKM01:Text-Parameter-Object-25:Value"] =
    function(x) 
      SetSmartMeterManufacturer(5, x)
    end,
}

------ local helper functions ------

--[[
Function to retrieve console output
]]
local function CaptureConsole(cmd, raw)
    local handle = assert(io.popen(cmd, 'r'))
    local output = assert(handle:read('*a'))
    
    handle:close()
    
    if raw then 
      return output 
    end
   
    output = string.gsub(
        string.gsub(
            string.gsub(output, '^%s+', ''), 
            '%s+$', 
            ''
        ), 
        '[\n\r]+',
        ' '
    )
   
   return output
end

--[[
Function to search an entry in a table with given key.
On found one entry returns the object and nil otherwise
]]
local function GetFromTable(_tbl, _key, ...)
  return not _key and _tbl or _tbl[_key] and GetFromTable(_tbl[_key], ...)
end

--[[
Function to search an entry in a table with given _key and _item.
On found one entry returns the object for _value and nil otherwise
]]
local function InTable(_tbl, _key, _item, _value)
  local searchValue = nil
  for key, value in pairs(_tbl) do
    if type(value) == 'table' then
      searchValue = InTable(value, _key, _item, _value)
    else
      if value == _item and key == _key then 
        return GetFromTable(_tbl, _value)
      end
    end
  end
  return searchValue
end

--[[
Function to get the enabled flag for a JSON object
]]
local function IsObjectEnabled(object)
  if object then
    local entry = GetFromTable(object, "enabled")
    if type(entry) == "string" then
      return (entry == "true")
    end
    return entry
  end
  return false
end

--[[
Function to get the KIM active flag for the JSON object data point
]]
local function IsObjectActivated(object)
  local metaData = GetFromTable(object, "metadata")
  if metaData
  then
    local metaEntry = InTable(metaData, "key", "KIM-IsActive", "value")
    if metaEntry then 
      return metaEntry == "1"
    end
  end
  return false
end

--[[
Function to get the enabled JSON object from API response or nill if no object was found
]]
local function GetObjectConfiguration(response)
  local j = json.decode(response)
  local object = GetFromTable(j, "response", "object")
  local isEnabled = IsObjectEnabled(object) or false
  if isEnabled then 
    return object
  end
  return nil
end

--[[
Function to get the object value from API response or nill if no value was found
]]
local function GetObjectValue(response)
  local j = json.decode(response)
  local value = GetFromTable(j, "response", "value")
  return value
end

--[[
Function to executes the defined functions in dpUrnToValueAction table for URNs in API response
]]
local function FillValues(response)
  local j = json.decode(response)
  local values = GetFromTable(j, "response", "values")
  
  local count = 0;
  for i,entry in ipairs(values) do
    local key = GetFromTable(entry, "urn") 
    local val = GetFromTable(entry, "value")
    if key and val then
      dpUrnToValueAction[key](val)
      count = count + 1
    end
  end
  return count
end

--[[ 
Function gets the table size
]]
local function GetTableSize(tbl)
  local getN = 0
  for n in pairs(tbl) do 
    getN = getN + 1 
  end
  return getN
end

------ public functions ------

-- device stack state parameter changed by test only
deviceStateParameter = {
  doWait = true,
  addPostMessage = false
}

--[[
Function to check the device stack state and wait 30s if state is not ready
]]
function HandleGetDeviceState(response)
  local j = json.decode(response)
  local devstate = GetFromTable(j, "response", "deviceState")
  if devstate then
    if type(devstate) == 'string' then
      if string.find(devstate, "ready") == nil then 
        if deviceStateParameter.doWait then 
          os.execute("sleep 30")
        end
      end
      if deviceStateParameter.addPostMessage then 
        return json.encode({devicestate  = devstate})
      end
      return "{}"
    end
  end
  return ""
end

-- this function is original from USC
function HandleGetDeviceConfig(response)
  local j = json.decode(response)
  local devcfg = GetFromTable(j, "response", "deviceConfig", "ipc")
  if devcfg
  then
    local fwv = GetFromTable(devcfg, "CurrentFirmwareVersion")
    local manu = GetFromTable(devcfg, "ManufacturerId")
    local device = GetFromTable(devcfg, "DeviceId")
    return json.encode({firmwareVersion = fwv, model = (manu .. device)})
  end
  return ""
end

--[[
Function to clears eCharge data for test only
]]
function ClearEChargeData()
  eChargeTable = {}
  gridServingTable = {}
  dlmTable = {}
  slmTable = {}
  chargePointsTable = {}
  applicationLoadState = true
  lmActivated = true
  dlmActivated = true
  slmActivated = true
  lmType = m.LMTYPE_DLM
end

--[[
Function to create eCharge data as json from tables
]]
function CreateEChargeData()
  eChargeTable[m.APPLICATION_LOAD_STATE_KEY_NAME] = applicationLoadState
  -- if application is unloaded, the state only is returned 
  if not applicationLoadState then
    return json.encode({eCharge = eChargeTable})
  end

  -- add data if table is not empty
  if not rawequal(next(gridServingTable), nil) then
    eChargeTable[m.GRID_SERVING_TABLE_NAME] = gridServingTable
  end
  if not rawequal(next(dlmTable), nil) then
    eChargeTable[m.DLM_TABLE_NAME] = dlmTable
  end
  if not rawequal(next(slmTable), nil) then
    eChargeTable[m.SLM_TABLE_NAME] = slmTable
  end
  
  -- add charge point data as array
  local key = m.CPS_TABLE_NAME
  for i,r in pairs(chargePointsTable) do
    if r then
      if eChargeTable[key] == nil then
        eChargeTable[key] = {}
      end
      table.insert(eChargeTable[key], r)
    end
  end

  -- if the table is empty, an empty json object is returned so no log entry is created
  if rawequal(next(eChargeTable), nil) then
    return "{}"
  end
  
  return json.encode({eCharge = eChargeTable})
end

--[[
Function to get the KNX Application program loaded state
]]
function HandleGetApplicationLoadStateConfiguration(response)
  local coObject = GetObjectConfiguration(response)
  if coObject then
      applicationLoadState = true
    else
      applicationLoadState = false
  end
  return "{}"
end

--[[
Function to get the grid serving interruption linked status from communication data point CO1
and add it to gridServingTable
]]
function HandleGetCo1Configuration(response)
  local coObject = GetObjectConfiguration(response)
  if coObject then
      gridServingTable[m.GRID_SERVING_TABLE_KEY_USED] = IsObjectActivated(coObject)
  end
  return "{}"
end

--[[
Function to get the DLM peak consumers linked status from communication data point CO13
and add it to DLM table
]]
function HandleGetCo13Configuration(response)
  -- if DLM is deactivated, an empty json object is returned so no log entry is created
  if not dlmActivated then 
    return "{}"
  end
  
  local object = GetObjectConfiguration(response)
  if object then
    SetLoadManagementTableRawValue(dlmTable, m.LMTYPE_DLM, m.LM_TABLE_PEAK_CONSUMERS_INDEX, IsObjectActivated(object))
  end
  return "{}"
end

--[[
Function to get data from values and add it to tables
]]
function HandleGetValues(response)
  if applicationLoadState then
    local count = FillValues(response)
  end
  return "{}"
end

--[[
Function to retrieve the uptime. Format <hour>:<minutes>:<seconds>
]]
function GetUptime()
  local uptimeValue = CaptureConsole("awk '{printf(\"%02d:%02d:%02d\",int($1/3600),int($1%3600/60),int($1%60))}' /proc/uptime", false)
  if uptimeValue then 
    return json.encode({uptime = uptimeValue}) 
  end
  
  return ""
end

return m
